#include "/lib/core.glsl"
#include "/lib/config.glsl"

layout(local_size_x = 32, local_size_y = 32, local_size_z = 1) in;
const vec2 workGroupsRender = vec2(1.0, 1.0);

uniform sampler2D depthtex0;
uniform layout(HDR_FMT) restrict writeonly image2D colorimg1;

uniform float day, far;
uniform vec2 pixSize;
uniform vec3 cameraPositionFract;
uniform mat4 gbufferModelViewInverse, gbufferProjectionInverse;
uniform sampler2D colortex1;
uniform usampler2D colortex2;

#include "/lib/fast_math.glsl"
#include "/lib/luma.glsl"
#include "/lib/octa_normal.glsl"

#if defined HAND_LIGHT && !defined LIGHT_LEVELS
	uniform float handLight;
#endif

#if COLORED_LIGHT || GLOSSY_REFLECTIONS
	uniform layout(rg32ui) restrict readonly uimage1D lightIndex;
	layout(shared, binding = 1) restrict readonly buffer lightIndexLen { int light_index; };

	shared uint sh_global_index_size;
	shared uint sh_index_size;
	shared ivec3 sh_bb_min;
	shared ivec3 sh_bb_max;
	shared uvec2[INDEX_SIZE] sh_index;
#endif

void main() {
	immut ivec2 texel = ivec2(gl_GlobalInvocationID.xy);
	immut float depth = texelFetch(depthtex0, texel, 0).r;
	immut vec2 coord = fma(gl_GlobalInvocationID.xy, pixSize, 0.5 * pixSize);
	immut vec4 view_undiv = gbufferProjectionInverse * vec4(fma(vec3(coord, depth), vec3(2.0), vec3(-1.0)), 1.0);
	immut vec3 plr = (gbufferModelViewInverse * vec4(view_undiv.xyz / view_undiv.w, 1.0)).xyz;

	#if COLORED_LIGHT || GLOSSY_REFLECTIONS
		if (gl_LocalInvocationIndex == 0u) {
			sh_global_index_size = light_index;
			sh_index_size = 0u;

			sh_bb_min = ivec3(INDEX_DIST);
			sh_bb_max = ivec3(-INDEX_DIST);
		}

		//if (anyInvocation(depth < 1.0)) {
			//memoryBarrierShared();
			barrier();

			if (length(plr) <= float(INDEX_DIST + 15)) {
				immut ivec3 i_plr = ivec3(fma(sign(plr).xxx, vec3(0.5), plr));

				atomicMin(sh_bb_min.x, i_plr.x); atomicMax(sh_bb_max.x, i_plr.x);
				atomicMin(sh_bb_min.y, i_plr.y); atomicMax(sh_bb_max.y, i_plr.y);
				atomicMin(sh_bb_min.z, i_plr.z); atomicMax(sh_bb_max.z, i_plr.z);
			}

			//memoryBarrierShared();
			barrier();

			immut vec3 bb_min = sh_bb_min;
			immut vec3 bb_max = sh_bb_max;

			if (all(greaterThanEqual(bb_max, bb_min))) {
				immut vec3 offset = -511.5 - cameraPositionFract;

				for (uint i = gl_LocalInvocationIndex; i < sh_global_index_size; i += gl_WorkGroupSize.x * gl_WorkGroupSize.y) {
					immut uvec2 light_data = imageLoad(lightIndex, int(i)).rg;

					immut vec3 pos = vec3(
						bitfieldExtract(light_data.x, 0, 10),
						bitfieldExtract(light_data.x, 10, 10),
						bitfieldExtract(light_data.x, 20, 10)
					) + offset;

					immut float emission = bitfieldExtract(light_data.y, 0, 4);

					// this may not be the best kind of culling but it works decently (world space bounding box)
					// it may be unneccessarily big when looking diagonally
					// something seems to be slightly wrong here as well
					if (all(greaterThanEqual(pos + emission, bb_min)) && all(lessThanEqual(pos - emission, bb_max))) {
						sh_index[atomicAdd(sh_index_size, 1u)] = light_data;
					}
				}
			}

			//memoryBarrierShared();
			barrier();
		//}
	#endif

	if (depth < 1.0) {
		immut uvec2 gbuffer_data = texelFetch(colortex2, texel, 0).rg;
		/* |-- R --|-- G --]
		   └8 8 8 8┴8 8 8 8┘
		    | | | | | | | |
		    | | | | | | └[light]: 2x8
		    | | | | └[normal]: 2x8
		    | | | └[smoothness]: 1x8
		    └[color]: 3x8
		*/

		vec3 color = texelFetch(colortex1, texel, 0).rgb;
		immut float dist = length(plr);

		immut vec4 normal_light = unpackUnorm4x8(gbuffer_data.y);
		vec2 light = normal_light.zw;
		light *= light;

		#ifdef NETHER
			const vec3 sky_light = vec3(0.3, 0.15, 0.2);
		#elif defined END
			const vec3 sky_light = vec3(0.15, 0.075, 0.2);
		#else
			immut float sky_light = 1.0 - lowp_sqrt(1.0 - max(0.1, light.y * clamp(day * 10.0, 0.25, 1.0))); // How did I even write this?
		#endif

		#ifdef LIGHT_LEVELS
			vec3 block_light = light.x > 0.0 ? mix(
				vec3(0.5, 0.25, 0.0),
				mix(
					vec3(0.75, 0.5, 0.0),
					mix(
						vec3(1.0, 0.75, 0.0),
						vec3(1.0, 1.0, 1.0),
						smoothstep(8.0, 15.0, light.x)
					),
					smoothstep(8.0, 11.0, light.x)
				),
				smoothstep(1.0, 7.0, light.x)
			) : vec3(0.5, 0.0, 0.0);
		#else
			#ifdef HAND_LIGHT
				light.x = mix(light.x, 1.0, max(0.0, handLight - dist * lowp_rcp(handLight * 16.0)));
			#endif

			vec3 block_light = light.xxx;
		#endif

		#if COLORED_LIGHT || GLOSSY_REFLECTIONS
			if (light.x >= float(MIN_INDEX_LL - 1) / 15.0 && dist < float(INDEX_DIST)) {
				vec3 shine = vec3(0.0);

				immut vec3 world_normal = normalize(octa_decode(normal_light.xy));
				immut float gb_smoothness = bitfieldExtract(gbuffer_data.x, 24, 8) / 256.0;
				immut float reflectance = gb_smoothness * gb_smoothness * float(SPECULAR * 10);
				immut float smoothness = fma(lowp_sqrt(gb_smoothness), 3.0, 1.0);

				immut vec3 n_plr = normalize(plr);
				immut vec3 offset = -511.5 - cameraPositionFract - plr;

				for(uint i = 0u; i < sh_index_size; ++i) {
					immut uvec2 light_data = sh_index[i];
					/* |--  R   --|--  G   --]
					   └10 10 10 2┴4 4 4 4 16┘
					    |  |  |  | | | | | └X
					    |  |  |  | | └[color]: 3x4
					    |  |  |  | └[emission]: 1x4
					    |  |  |  └X
					    └[ pos ]: 3x10
					*/

					immut vec3 pos = vec3(
						bitfieldExtract(light_data.x, 0, 10),
						bitfieldExtract(light_data.x, 10, 10),
						bitfieldExtract(light_data.x, 20, 10)
					) + offset;

					immut float pos_len = length(pos);
					immut float dist = pos_len - 1.0;
					immut float emission = bitfieldExtract(light_data.y, 0, 4);

					if (dist <= emission) {
						immut vec3 n_pos = pos * lowp_rcp(pos_len);
						immut float lambertian = dot(world_normal, n_pos);
						immut float lighting = lambertian > 0.01 ? fma(pow(max(dot(n_plr, n_pos), 0.0), smoothness), reflectance, lambertian) : IND_ILLUM;

						shine = fma(vec3(
							bitfieldExtract(light_data.y, 4, 4),
							bitfieldExtract(light_data.y, 8, 4),
							bitfieldExtract(light_data.y, 12, 4)
						), (lighting * (emission - dist)).xxx, shine);
					}
				}

				if (dot(shine, vec3(1.0)) > 0.01) {
					shine /= 225.0; // Undo the multiplication from packing light color and emission

					vec3 new_light = block_light;

					#if COLORED_LIGHT
						immut vec3 color = normalize(shine);

						new_light = mix(new_light, new_light * color * lowp_rcp(luminance(color)), COLORED_LIGHT * 0.1);
					#endif

					#if GLOSSY_REFLECTIONS
						new_light = fma(light.x * shine, vec3(float(GLOSSY_REFLECTIONS) * 0.5), new_light);
					#endif

					block_light = mix(block_light, new_light, (1.0 - smoothstep(float(INDEX_DIST - 15), float(INDEX_DIST), dist)));
				}
			}

			// debug culling
			// color.gb += vec2(sh_index_size < sh_global_index_size, sh_index_size == 0);
		#endif

		imageStore(colorimg1, texel, vec4(
			color * fma(block_light, vec3(1.0, 0.8, 0.7) * lowp_rcp(1.0 + sky_light), vec3(sky_light)),
			0.0
		));
	}
}