#include "/lib/core.glsl"
#include "/lib/config.glsl"

uniform vec3 shadowLightPosition;
uniform mat3 normalMatrix;
uniform mat4 modelViewMatrix, projectionMatrix, shadowModelView, shadowProjection, textureMatrix;

#if COLORED_LIGHT || GLOSSY_REFLECTIONS
	layout(shared, binding = 1) restrict readonly buffer lightIndexLen { int light_index; };
#endif

#include "/lib/fast_math.glsl"
#include "/lib/distort.glsl"
#include "/lib/vertex_sample_index.glsl"

#ifdef TERRAIN
	uniform vec3 chunkOffset;

	in vec2 mc_Entity;
	in vec4 at_midBlock;

	#if WAVES
		#include "/lib/wave.glsl"
	#endif
#endif

#ifdef ENTITY_COLOR
	uniform vec4 entityColor;
#endif

#ifndef NO_NORMAL
	in vec3 vaNormal;
#endif

in vec2 vaUV0;
in vec3 vaPosition;
in vec4 vaColor;

out VertexData {
	vec2 coord;
	vec3 s_ndc, view;

	#ifndef NO_NORMAL
		vec3 normal;
	#endif

	#ifdef TERRAIN
		vec4 color;
	#else
		vec3 color;
	#endif
} f;

void main() {
	immut vec2 coord = (textureMatrix * vec4(vaUV0, 0.0, 1.0)).xy;
	vec3 model = vaPosition;

	#ifdef TERRAIN
		vec4 color = vec4(vaColor.rgb, 1.0);

		model += chunkOffset;

		if (mc_Entity.y > 0.5) {
			color.a *= WATER_OPACITY * 0.01;

			#if WAVES
				model.y += wave(model.xz);
			#endif
		}
	#else
		vec3 color = vaColor.rgb;
	#endif

	#ifdef ENTITY_COLOR
		color.rgb = mix(color.rgb, entityColor.rgb, entityColor.a);
	#endif

	immut vec4 model_4 = vec4(model, 1.0);
	immut vec4 view = modelViewMatrix * model_4;
	vec3 s_clip = (shadowProjection * (shadowModelView * model_4)).xyz;

	immut vec3 normal = normalMatrix * 
		#ifdef NO_NORMAL
			vec3(0.0, -1.0, 0.0);
		#else
			vaNormal;
		#endif

	vec3 light = get_fancy_light(view.xyz, normal);

	immut float lambertian = dot(normal, shadowLightPosition * 0.01);

	s_clip = distort(s_clip);
	float s_bias = max(0.00005, 0.0003 * length(s_clip.xy)) * lowp_rcp(lambertian);

	#ifdef TERRAIN
		#if SSS // todo!() check that this makes sense
			if (mc_Entity.x == 0.0 && lambertian < 0.0) s_bias *= -1.0;
		#endif

		immut float emission = clamp((mc_Entity.x + at_midBlock.w) / 15.0, 0.0, 1.0);
		light = max(light, vec3(emission)) + emission * 0.3;
	#endif

	s_clip.z -= s_bias;
	f.s_ndc = s_clip * 0.5 + 0.5;

	color.rgb *= light;

	f.coord = (textureMatrix * vec4(coord, 0.0, 1.0)).xy;
	f.color = color;
	f.view = view.xyz;

	#ifndef NO_NORMAL
		f.normal = normal;
	#endif

	gl_Position = projectionMatrix * view;
}